;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;;  splat-fluent.scm
;;;  Class definition and methods for fluents and predicated actions
;;;  
;;;  CVS Identifier:
;;;  $Id$
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;


;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;;  <splat-predicated-action>
;;;  The predicate-thunk is a function which is evaluated any time the 
;;;  relevant-variables change value.  If it returns non-#f, the
;;;  consequent-thunk is evaluated. 
;;;
;;;  The predicated-action must be enabled with (enable var)
;;;  before it does anything.  The default is only to take the 
;;;  action once, automatically disabling before the consequent is 
;;;  called.
;;;  
;;;  It is not permitted to call enable from within the consequent.
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

(define-class <splat-predicated-action> (<object>)
  (predicate-thunk    init-value: #f
                      init-keyword: predicate:)
  (consequent-thunk   init-value: #f
                      init-keyword: action:)
  (enabled?           init-value: #t init-keyword: #f)
  (relevant-variables init-value: '() init-keyword: fluents:)
  (mutex              init-value: #f init-keyword: #f))


;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;;  <splat-fluent>
;;;  This is a notify-on-change variable that works with 
;;;  <splat-predicated-action>.  The binding must be accessed only
;;;  with get-fluent and set-fluent!.  
;;;
;;;  set-fluent!  will make an equivalence test with equiv-thunk 
;;;  and won't call the predicate-thunk if it is true.
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

(define-class <splat-fluent> (<object>)
  (binding      init-value:   #f  
                init-keyword: binding:)
  (equiv-thunk  init-value:   #f
                init-keyword: equiv-thunk:)
  (mutex        init-value:   #f
                init-keyword: #f)
  (actions      init-value:   '()
                init-keyword: #f))

;;; methods for <splat-fluent>
(define-generic-function set-fluent!)
(define-generic-function get-fluent)
(define-generic-function add-predicated-action!)
(define-generic-function remove-predicated-action!)

;;; methods for <splat-predicated-action> 
(define-generic-function predicate?)
(define-generic-function consequent)
(define-generic-function enable)
(define-generic-function disable)


;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;;  <splat-fluent> methods 
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

(define-method initialize ((self <splat-fluent>) . other-args)
  (set-mutex! self (make <mutex>)))


;;;  modify the binding of the fluent, checking any dependent 
;;;  splat-predicated-actions if appropriate. 
(define-method set-fluent! ((self <splat-fluent>) new-value)
  (semaphore-wait (mutex self))
  (let ((old-value (binding self))
        (need-update #t))
    (set-binding! self new-value)
    (semaphore-signal (mutex self))
    
    ;; if there's an equiv-thunk, call it and avoid the predicate
    ;; calls if the values are equivalent.
    (if (procedure? (equiv-thunk self))
        (set! need-update (not ((equiv-thunk self) old-value new-value))))    
    (if need-update
        (for-each (lambda (elt)              
                    (if (predicate? elt)
                        (begin 
                          (disable elt)
                          (consequent elt))))
                  (actions self)))
    new-value))


;;;  get the binding of the fluent.
(define-method get-fluent ((self <splat-fluent>))
  (let ((retval #f))
    (semaphore-wait (mutex self))
    (set! retval (binding self))
    (semaphore-signal (mutex self))
    retval))

;;;  add a predicated action to the notify list of the fluent.
(define-method add-predicated-action! ((self <splat-fluent>)
                                       (action <splat-predicated-action>))
  (semaphore-wait (mutex self))
  (set-actions! self (cons action (actions self)))
  (semaphore-signal (mutex self)))

;;;  take a predicated action off the notify list of the fluent.
(define-method remove-predicated-action! ((self <splat-fluent>)
                                          (action <splat-predicated-action>))
  (semaphore-wait (mutex self))
  (set-actions! self (do ((list-left (actions self) (cdr list-left))
                          (new-list '()))       
                         ((null? list-left) new-list)
                       (if (not (eq? (car list-left) action))
                           (set! new-list (cons (car list-left) new-list)))))
  (semaphore-signal (mutex self)))


;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;;  <splat-predicated-action> methods
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

(define-method initialize ((self <splat-predicated-action>) . other-args)
  (set-mutex! self (make <mutex>))
  (if (not (procedure? (predicate-thunk self)))
      (set-predicate-thunk! self (lambda (self) #t)))
  (if (not (procedure? (consequent-thunk self)))
      (set-consequent-thunk! self (lambda (self) #t))))

(define-method predicate? ((self <splat-predicated-action>))
  ((predicate-thunk self) self))

(define-method consequent ((self <splat-predicated-action>)) 
  ((consequent-thunk self) self))

;;;  if the predicated-action is enabled with the condition true,
;;;  the consequent is run immediately. 
(define-method enable ((self <splat-predicated-action>))
  (semaphore-wait (mutex self))
  (if (predicate? self)			
      (begin 
        (consequent self)			
        (set-enabled?! self #f))
      (begin
        (for-each (lambda (variable)
                    (add-predicated-action! variable self))
                  (relevant-variables self))
        (set-enabled?! self #t)))
  (semaphore-signal (mutex self)))

;;; disable 
(define-method disable ((self <splat-predicated-action>))
  (semaphore-wait (mutex self))
  (for-each (lambda (variable) 
              (remove-predicated-action! variable self))
            (relevant-variables self))
  (set-enabled?! self #f)
  (semaphore-signal (mutex self)))

(values #t)
