;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;;  splat-syntax.scm
;;;  Rewriters for defining splats and doing useful things. 
;;;  
;;;  CVS Identifier:
;;;  $Id$
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

(define (define-splat-error exit-proc . args)
  (format #t ";; define-splat: error: ")
  (apply format `(#t ,@args))
  (exit-proc #f))

(define (define-behavior-error exit-proc . args)
  (format #t ";; define-behavior: error: ")
  (apply format `(#t ,@args))
  (exit-proc #f))

(define (define-splat-warning . args)
  (format #t ";; define-splat: warning: ")
  (apply format `(#t ,@args)))


;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;;  (run splat)
;;;  Run a single splat and wait for it.  This can ONLY be used where
;;;  self is in scope, i.e. inside a splat definition.
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

(define-rewriter (run form)
  `(begin
     (let ((spl ,(cadr form)))
       (start-task spl self)
       (wait-for-task spl self))))


;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;;  (sequence-actions a1 .. an)
;;;  Launch a sequence of splats and wait for each to complete. 
;;;  This can only be used within a splat. 
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

(define-rewriter (sequence-actions form)
  (let ((action-sequence '())
        (rewritten-form #f))
    ;; iterate over the splats, building the code to launch and 
    ;; wait for each one.
    (let loop ((splats-remaining (cdr form))
               (current-splat (cadr form)))
      (set! action-sequence
            (append action-sequence
                    `((begin (set! temp ,current-splat)
                             (if (splat? temp)
                                 (begin
                                   (start-task temp self)
                                   (set! rval (wait-for-task temp self)))
                                 (set! rval temp))
                             (if (not rval)
                                 (exit-proc (list #f (name temp)))
                                 rval)))))
      (set! splats-remaining (cdr splats-remaining))
      (if (not (null? splats-remaining))
          (begin
            (set! current-splat (car splats-remaining))
            (loop splats-remaining current-splat))))
    
    ;; wrap that list up as the only method of a temporary splat,
    ;; which is what gets returned from the sequence-actions call.
    ;; Just like any other splat, the result must be run to do anything.
    ;; the call/cc lets any failing splat bail out in an and-like way. 
    (set! rewritten-form 
          `(let ((tempsplat (make <splat>)))             
             (add-method! 
              tempsplat 
              (make 
               <splat-method> context: (lambda () #t)  action: 
               (lambda ()
                 (let ((retval
                        (call-with-current-continuation 
                         (lambda (exit-proc) 
                           (let ((temp #f) (rval #f))
                             ,@action-sequence)))))
                   (finish tempsplat retval)))))
             tempsplat))
    rewritten-form))



;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;;  (parallel-actions form)
;;;  Launch several splats in parallel and wait for all to complete.
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

(define-rewriter (parallel-actions form)
  (format #t ";;; parallel-actions compiling\n")
  (let ((action-sequence '())
        (rewritten-form #f))
    (let loop ((splats-remaining (cdr form))
               (current-splat (cadr form)))
      ;; at compile time, build action-sequence which is the form
      ;; containing the splat launching code.  At run time, 
      ;; build task-list, which is the list of launched splats which
      ;; must be waited on.
      (set! action-sequence
            (append action-sequence
                    `((begin (set! temp ,current-splat)
                             (if (splat? temp)
                                 (begin
                                   (set! task-list 
                                         (append task-list (list temp)))
                                   (start-task temp tempsplat)))))))
      (set! splats-remaining (cdr splats-remaining))
      (if (not (null? splats-remaining))
          (begin
            (set! current-splat (car splats-remaining))
            (loop splats-remaining current-splat))))
    
    ;; wrap the action-sequence as the only method of a temp splat.
    ;; the wait-for-task calls wait in the order that the splats were 
    ;; launched.  Shouldn't matter, should it?
    (set! rewritten-form 
          `(let ((tempsplat (make <splat>)))             
             (add-method! 
              tempsplat 
              (make 
               <splat-method> context: (lambda () #t)  action: 
               (lambda ()
                 (let ((retval
                        (let ((temp #f) (rval #f) (task-list '()))
                          ,@action-sequence
                          (do ((list-left task-list (cdr list-left)))
                              ((null? list-left) rval)
                            (set! temp (car list-left))
                            (set! rval (wait-for-task temp tempsplat))))))
                   (finish tempsplat retval)))))
             tempsplat))
    rewritten-form))


;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;;  (define-behavior form)
;;;  Simple syntax form to define a behavior.  Behaviors don't have 
;;;  discrete-action nature, can be modulated by some undefined 
;;;  mechanism, and can be turned on and off as needed. 
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

(define-rewriter (define-behavior form)
  (call-with-current-continuation
   (lambda (break)
     (let ((behavior-name #f)
           (argument-list #f)
           (behavior-vars #f)
           (form-list #f)
           (this-form #f)
           (rewritten-form '()))
       (format #t ";; define-behavior: compiling behavior...~%")
       
       ;; check for format errors 
       (if (not (list? form))
           (define-behavior-error break "define-rewriter got a bad form.~%"))
       (if (null? (cdr form))
           (define-behavior-error break "no behavior specified.~%") )
       (if (not (list? (cadr form)))
           (define-behavior-error break "bad binding form.~%"))
       (if (not (symbol? (caadr form)))
           (define-behavior-error break "bad name for behavior.~%"))
       (if (null? (cddr form))
           (define-behavior-error break "empty behavior body.~%"))
       
       (set! form-list (cddr form))
       (set! this-form (car form-list))
       (set! behavior-name (caadr form))
       (set! argument-list (cdadr form))
       
       (format #t ";; define-behavior: name = ~s~%" behavior-name)
       (format #t ";;                  args = ~s~%" argument-list)
       (format #t ";; define-behavior: compiling ~s component forms~%"
               (length form-list))
       
       (let loop ()
         (set! temp (cond ((eq? (car this-form) 'enable)
                           `(enable-thunk: (lambda () 
                                             ,@(cdr this-form))))
                          ((eq? (car this-form) 'disable)
                           `(disable-thunk: (lambda () 
                                              ,@(cdr this-form))))
                          ((eq? (car this-form) 'control)
                           `(control-thunk: (lambda () 
                                              ,@(cdr this-form))))
                          ((eq? (car this-form) 'local-vars)
                           (cdr this-form))))
         (if (eq? (car this-form) 'local-vars)
             (set! behavior-vars temp)
             (set! rewritten-form (append rewritten-form temp)))
         
         (set! form-list (cdr form-list))
         (if (not (null? form-list))
             (begin 
               (set! this-form (car form-list))
               (loop))))
       (format #t ";; define-behavior: finished compiling ~s~%"
               behavior-name)
       (set! rewritten-form 
             `(define ,behavior-name 
                (lambda (,@argument-list)
                  (let* ((self #f) ,@behavior-vars)
                    (set! self (make <splat-behavior> 
                                     ,@rewritten-form))
                    self))))
;       (format #t "~%Rewritten form:~%~s~%" rewritten-form)
       rewritten-form))))
               

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;;  (define-splat form) 
;;;  The main syntax form for defining a type of splat.
;;;
;;;  The value returned is a thunk which can be called to return a 
;;;  splat data structure.
;;;
;;;  All the ancillary lambdas defined by this rewriter have the 
;;;  argument-list in their scope, in addition to (self <splat>),
;;;  and are called that way. 
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

(define-rewriter (define-splat form)
  (call-with-current-continuation
   (lambda (break)      
     (let ((splat-name #f)
           (argument-list #f)
           (current-form #f)
           (form-list #f)
           (assembled-args #f)
           (rewritten-form #f)
           (other-splat-vars '())
           (temp #f))

       ;; print a friendly status message. 
       (format #t ";; define-splat: compiling SPLAT...~%")
       
       (if (not (list? form))
           (define-splat-error break "define-rewriter got a bad form.~%"))
       (if (null? (cdr form))
           (define-splat-error break "no SPLAT specified.~%") )
       (if (not (list? (cadr form)))
           (define-splat-error break "bad binding form.~%"))
       (if (not (symbol? (caadr form)))
           (define-splat-error break "bad name for SPLAT.~%"))
       (if (null? (cddr form))
           (define-splat-error break "empty SPLAT body.~%"))
       
       (set! form-list (cddr form))
       (set! splat-name (caadr form))
       (set! argument-list (cdadr form))
       (set! rewritten-form '())

       (format #t ";; define-splat: name = ~s~%" splat-name)
       (format #t ";;               args = ~s~%" argument-list)
       (format #t ";; define-splat: compiling ~s SPLAT component forms~%" 
               (length form-list))

       ;; loop over all the main forms in the define-splat body.
       (let continue ()
         
         ;; get the next form from the defsplat 
         (set! current-form (car form-list))
         (set! form-list (cdr form-list))
         
         ;; error checking 
         (if (or (not (list?   current-form))
                 (not (symbol? (car current-form))))
             (define-splat-error break "bad section type ~s~%" current-form))
         
         ;; figure out what kind it is and tack the appropriate 
         ;; rewritten bits on to the current form 
         (set! temp 
               (cond 
                ((eq? (car current-form) 'local-vars)
                 (list (local-vars-rewriter current-form break)))
                ((eq? (car current-form) 'preconditions)
                 (list (preconditions-rewriter current-form break)))
                ((eq? (car current-form) 'initializer)
                 (list (initializer-rewriter current-form break)))
                ((eq? (car current-form) 'finalizer)
                 (list (initializer-rewriter current-form break)))
                ((eq? (car current-form) 'event-handler)
                 (list (event-handler-rewriter current-form break)))
                ((eq? (car current-form) 'wait-for)
                 (list (wait-for-rewriter current-form break)))
                ((eq? (car current-form) 'method)
                 (list (method-rewriter current-form break)))
                (#t 
                 (define-splat-error break "unknown section type ~s~%"
                   (car current-form)))))

         (if (eq? (car current-form) 'local-vars)
             (set! other-splat-vars (car temp))
             (set! rewritten-form (append rewritten-form temp)))
         
         (if (null? form-list)
             (begin 
               (format #t ";; define-splat: finished compiling SPLAT ~s~%" 
                       splat-name))
             (continue)))
       
       ;; now we have successfully finished the form rewriting. 
       ; (set! rewritten-form `(let ((self (make <splat> name: ,splat-name)))))

       (set! rewritten-form 
             `(define ,splat-name 
                (lambda (,@argument-list)
                  (let* ((self (make <splat> name: (quote ,splat-name)))
                        ,@other-splat-vars)
                    ,@rewritten-form self))))
       
;       (format #t "~%Rewritten form:~%~s~%" rewritten-form)
       rewritten-form))))

(define (local-vars-rewriter form exit-proc)
  (format #t ";; define-splat: compiling (local-vars ...)~%")
  (let ((var-list (cdr form))
        (rest-list (cdr form))
        (this-elt #f))    
    (if (null? var-list)
        (define-splat-warning "empty local-var binding form.~%"))
    (let loop () 
      (set! this-elt (car rest-list))
      (set! rest-list (cdr rest-list))
      (if (or (not (list? this-elt))
              (not (eq? (length this-elt) 2)))
          (define-splat-error exit-proc "bad local-var binding form.~%"))
      (if (not (null? rest-list))
          (loop)))
    var-list))


(define (preconditions-rewriter form exit-proc)
  (format #t ";; define-splat: compiling (precondition ...)~%")

  (let ((precond-list (cdr form))
        (this-precond #f)
        (rewritten-form '(and)))
    ;; check for a null precondition list that would kill (and )
    (if (null? precond-list)
        (begin
          (define-splat-warning "empty precondition form.~%")
          (set! precond-list '(#t))))
    (do ()
        ((null? precond-list) #t)
      (set! this-precond (car precond-list))
      (set! precond-list (cdr precond-list))
      (set! rewritten-form 
            (append rewritten-form
                    (list this-precond))))
    `(add-precondition! self (lambda () ,rewritten-form))))

(define (initializer-rewriter form exit-proc)
  (format #t ";; define-splat: compiling (initializer ...)~%")
  
  (let ((initializer-list (cdr form))
        (this-initializer #f)
        (rewritten-form '(lambda ())))
    (if (null? initializer-list)
        (begin 
          (define-splat-warning "empty initializer form~%")
          (set! initializer-list '(#t))))
    (do ()
        ((null? initializer-list) #t)
      (set! this-initializer (car initializer-list))
      (set! initializer-list (cdr initializer-list))
      (set! rewritten-form 
            (append rewritten-form
                    (list this-initializer))))
    `(add-initializer! self ,rewritten-form)))

(define (finalizer-rewriter form exit-proc)
  (format #t ";; define-splat: compiling (finalizer ...)~%")

  (let ((finalizer-list (cdr form))
        (this-finalizer #f)
        (rewritten-form '(lambda ())))
    (if (null? finalizer-list)
        (begin 
          (define-splat-warning "empty finalizer form~%")
          (set! finalizer-list '(#t))))
    (do ()
        ((null? finalizer-list) #t)
      (set! this-finalizer (car finalizer-list))
      (set! finalizer-list (cdr finalizer-list))
      (set! rewritten-form 
            (append rewritten-form
                    (list this-finalizer))))
    `(add-finalizer! self ,rewritten-form)))

(define (event-handler-rewriter form exit-proc)
  (format #t ";; define-splat: compiling (event-handler ...)~%")
  ;; event-handlers specify a list of splat-fluents, a predicate,
  ;; and an action.
  (if (not (eq? (length (cdr form)) 2))
      (define-splat-error exit-proc
        "event-handler form requires exactly 2 arguments, given ~s~%"
        (length (cdr form))))
  (let* ((event-spec #f)
         (action-spec #f)
         (rewritten-form #f))

    ;; assign an event-spec and an action-spec 
    (cond ((and (eq? (caadr form) 'event)
                (eq? (caaddr form) 'action))
           (begin
             (set! event-spec (cadr form))
             (set! action-spec (caddr form))))
          ((and (eq? (caadr form) 'action)
                (eq? (caaddr form) 'event))
           (begin
             (set! action-spec (cadr form))
             (set! event-spec (caddr form))))
          (#t
           (define-splat-error exit-proc 
             "event-handler requires one event form and one action form.~%")))

    ;; check for errors 
    (if (or 
         (null? (cdr event-spec))
         (not (eq? 2 (length (cdr event-spec))))
         (not (list? (cadr event-spec)))
         (null? (cddr event-spec)))              
        (define-splat-error exit-proc
          "event-specifier must contain a list of fluents and a predicate~%"))
    (if (null? (cdr action-spec))
        (define-splat-error exit-proc
          "action-specifier must contain one or more action forms.~%"))
    
    (set! rewritten-form 
          `(add-event-handler! self 
                               (make <splat-predicated-action> 
                                     fluents: 
                                     (list ,@(cadr event-spec))
                                     predicate: (lambda ()
                                                  ,@(cddr event-spec))
                                     action: (lambda ()
                                               ,@(cdr action-spec)))))
    rewritten-form))

(define (wait-for-rewriter form exit-proc)
  (format #t ";; define-splat: compiling (wait-for ...)~%")
  ;; a wait-for is a splat-predicated-action that unblocks the 
  ;; splat when the condition becomes true.
  (if (or (null? (cdr form))
          (not (eq? 2 (length (cdr form))))
          (not (list? (cadr form)))
          (null? (cadr form)))      
      (define-splat-error exit-proc 
        "wait-for requires a list of fluents and a predicate~%"))
  
  (let* ((fluent-list (cadr form))
         (predicate-forms (cddr form))
         (rewritten-form #f))
    (set! rewritten-form 
          `(add-wait-pred! 
            self 
            (make <splat-predicated-action> 
                  fluents: (list ,@fluent-list)
                  predicate: (lambda ()
                               ,@predicate-forms)
                  action: #f)))
    rewritten-form))

(define (method-rewriter form exit-proc)
  (format #t ";; define-splat: compiling (method ...)~%")
  ;; this one's pretty straightforward 
  (let ((rewritten-form #f)
        (method-context #f)
        (method-action #f))
    (if (or (null? (cdr form))
            (> 2 (length (cdr form))))
        (define-splat-error exit-proc
          "method requires a context form and an action body.~%"))
    (set! method-context (cadr form))
    (set! method-action (cddr form))
    (set! rewritten-form 
          `(add-method! self 
                        (make <splat-method> 
                              context: (lambda () ,method-context)
                              action: 
                              (lambda ()
                                ,@method-action ))))))


(values #t)