;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;;  splat-thread.scm
;;;  The thread function which is a running splat.
;;;  
;;;  CVS Identifier:
;;;  $Id$
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;


;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;; utility functions for the splat thread 
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

(define (check-preconditions (self <splat>))
  (let ((precond-all-true #t))
    (for-each (lambda (this-predicate) 
                (set! precond-all-true 
                      (and precond-all-true 
                           (this-predicate))))
              (precond-list self))
    precond-all-true))


(define (run-initializers (self <splat>))
  (set-initializers-run! self #t)
  (for-each (lambda (this-thunk)
              (this-thunk))
            (initializer-list self)))


(define (do-wait-preds (self <splat>))
  (let* ((wait-preds-started 0)
         (wait-preds-finished 0)
         (wait-pred-mutex (make <mutex>))
         (wait-pred-condition (make <condition-variable>)))
    
    ;; enable all the predicated-actions 
    (for-each 
     (lambda (this-wait-pred)
       (let ((old-action (consequent-thunk this-wait-pred)))
         (semaphore-wait wait-pred-mutex)
         (set! wait-preds-started (+ 1 wait-preds-started))
         (semaphore-signal wait-pred-mutex)

;         (with-semaphore
;          wait-pred-mutex
;          (set! wait-preds-started (+ 1 wait-preds-started)))
         
         (set-consequent-thunk! this-wait-pred 
                                (lambda () 
                                  (old-action this-wait-pred)
                                  (semaphore-wait wait-pred-mutex)
                                  (set! wait-preds-finished 
                                        (+ 1 wait-preds-finished))
                                  (semaphore-signal wait-pred-mutex)
                                  (condition-broadcast wait-pred-condition)))
         (enable this-wait-pred)))
     (wait-pred-list self))
    
    ;; wait for wait-preds to become true 
    (semaphore-wait wait-pred-mutex)
    (do ()
        ((eq? wait-preds-started wait-preds-finished) #t)
      (condition-wait wait-pred-condition wait-pred-mutex))
    (semaphore-signal wait-pred-mutex)))


(define (install-event-handlers (self <splat>))
  (for-each (lambda (handler) 
              (if (not (enabled? handler))
                  (enable handler)))
            (event-handler-list self)))


(define (uninstall-event-handlers (self <splat>))
  (for-each (lambda (handler) 
              (if (enabled? handler)
                  (disable handler)))
            (event-handler-list self)))


(define (select-method (self <splat>))
  (do ((methods-left (method-list self) (cdr methods-left))
       (method-to-run #f))              
      ((or (null? methods-left) method-to-run) method-to-run)
    (if (and (not (method-tried (car methods-left)))
             ((context (car methods-left))))
        (begin 
          (set! method-to-run (car methods-left))
          (set! methods-left '(#t))))))

(define (run-method (self <splat>) (method <splat-method>))
  (let ((retval #f))
    (set! retval (call-with-current-continuation 
                  (lambda (exit-c) 
                    (set-exit-proc! self exit-c)
                    (execute method self))))))

(define (run-finalizers (self <splat>))
  (for-each (lambda (this-thunk)
              (this-thunk self))
            (finalizer-list self)))


;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;;  eval-splat-thread
;;;  the splat thread function. 
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

(define-method eval-splat-thread ((self <splat>) . other-args)
  (set-state! self 'running)
  
  ;; iterate over method selection until preconditions fail or 
  ;; a method succeeds.
  (call-with-current-continuation
   (lambda (break)
     (let loop () 

       ;; exit the loop when state is finished 
       (if (eq? (state self) 'finished)
           (break (return-value self)))
       
       ;; if preconditions ever fail, stop trying to select methods. 
       (if (not (check-preconditions self))
           (begin 
             (format #t ";;;  preconditions not true. exiting.\n")
             (with-semaphore (mutex self)
                             (set-state! self 'finished)
                             (set-return-value! self 'failed-precond))
             (condition-broadcast (condition self))
             (loop)))
       
       ;; preconditions are true.  Run the initializers.
       (run-initializers self)
       
       ;; install and wait for the wait predicates 
       (do-wait-preds self)
       
       ;; install event handlers 
       (install-event-handlers self)
       
       ;; select a method 
       (let ((selected-method (select-method self)))
         (if (not selected-method)
             (begin 
               (format #t ";;;  no applicable methods.  exiting.\n")
               (with-semaphore (mutex self)
                               (set-state! self 'finished))
               (condition-broadcast (condition self))
               (set-return-value! self 'no-method)
               (loop)))
         (run-method self selected-method)
         (if (eq? (return-value self) 'method-succeeded)
             ;;  method succeeded.  Yay! 
             (begin 
               (format #t ";;;  method succeeded in ~s\n" (name self))
               (with-semaphore (mutex self)
                               (set-state! self 'finished))
               (condition-broadcast (condition self))
               (uninstall-event-handlers self)
               (set-return-value! self 'method-succeeded)
;               (format #t ";;;   about to break ~s\n" (name self))
               (break #t))
             ;;  method failed.  Just loop.
             (begin 
               (format #t ";;;  method failed... selecting another in ~s\n"
                       (name self))
               (loop)))))))
  
  ;; clean up
  (if (initializers-run self) 
      (run-finalizers self))
  (format #t ";;;   finishing thread in ~s~%" (name self)))

(values #t)

